; Author: David Bennett - Version 1.0 - Date: 11/9/88
;
; Version 1.1 - Added simple windowing routines, including clipping at
;               borders.  THomas Hill, Jan 1989
;
; This is a module of direct video writing routines for TASM.  Please see the
; file VIDEO.DOC for more information.  Examples of most routines are set forth
; in VIDDEMO.ASM.
;
; -Dave
;
; Jan 24, 1989 - Several changes had to be made in order to use my windowing
;                with the following video routines.  In particular, the 
;                screen saving and restoring procedures were not set up to
;                accept storage to/from diffrent data segments.  I changed
;                the parameters passed to include the segment of the stored
;                data.
;      
;                Thomas Hill
;
; Jan 29, 1994 - Taken out snow-checking stuff as the majority of PC owners
;                now use EGA/VGA graphics and procs that I don't need.
;                Modified to interface with C programs. Please see CONIO.C
;                for examples.
;
;                Hung Vu
;

.MODEL small,C
.DATA
EXTRN BaseOfScreen:WORD
.CODE
MoveXY_DI       PROC
	; This procedure moves to the offset indicated by an X & Y cusor
	; location.
	; Input
	;       AH = Row
	;       AL = Column
	; Output
	;       DI = Memory Offset
		push    cx              ; Save CX
		xor     cl, cl          ; Clear CL
		mov     ch, ah          ; CX = Row * 256
		dec     ch              ; CX = (Row - 1) {0-24 Based}
		shr     cx, 1           ; CX = Row * 128
		mov     di, cx          ; Store in DI
		shr     di, 1           ; DI = Row * 64
		shr     di, 1           ; DI = Row * 32
		add     di, cx          ; DI = (Row * 128)+(Row * 32) {Row*160}
		xor     ch, ch          ; Clear CH register
		mov     cl, al          ; CX = Columns
		dec     cx              ; Make 0-79
		shl     cx, 1           ; Account for attribute
		add     di, cx          ; DI = (Row * 160) + (Col * 2)
		pop     cx              ; Restore CX register
		ret
MoveXY_DI       ENDP

MoveXY_SI       PROC
	; This procedure moves to the offset indicated by an X & Y cusor
	; location.
	; Input
	;       AH = Row
	;       AL = Column
	; Output
	;       SI = Memory Offset - Points 1 byte beyond null of str displayed
		push    cx              ; Save CX
		xor     cl, cl          ; Clear CL
		mov     ch, ah          ; CX = Row * 256
		dec     ch              ; CX = (Row - 1) {0-24 Based}
		shr     cx, 1           ; CX = Row * 128
		mov     si, cx          ; Store in SI
		shr     si, 1           ; SI = Row * 64
		shr     si, 1           ; SI = Row * 32
		add     si, cx          ; SI = (Row * 128)+(Row * 32) {Row*160}
		xor     ch, ch          ; Clear CH register
		mov     cl, al          ; CX = Columns
		dec     cx              ; Make 0-79
		shl     cx, 1           ; Account for attribute
		add     si, cx          ; DI = (Row * 160) + (Col * 2)
		pop     cx              ; Restore CX register
		ret
MoveXY_SI       ENDP

                PUBLIC DWriteCH
DWriteCH        PROC C
	; Writes a character to the screen using direct memory access.
        ; *** IMPORTANT -- CALL InitVideo Before using this routine!
	; Input
	;       AH      Row on screen 1-25
	;       AL      Column on screen 1-80
	;       BH      Video Attribute
	;       BL      Character
	;       CX      Number of times
	; Output
	;       Screen memory (B000:0000 or 8000 CGA/MONO
		push    ax              ; Store the registers
		push    cx
		push    dx
		push    es
		push    di
		call    MoveXY_DI       ; Set DI to row / column offset
		mov     es, [BaseOfScreen]      ; Move screen seg to ES
		mov     ax, bx          ; Move char/attr into AX
		rep     stosw           ; Move char/attr to screen CX times
		pop     di              ; Restore the registers
		pop     es
		pop     dx
		pop     cx
		pop     ax
		ret
DWriteCH        ENDP

                PUBLIC DWriteStr
DWriteStr       PROC C
	; This procedure writes a null delimited string to the screen using
	; direct memory access.
        ; *** IMPORTANT -- CALL InitVideo Before using this routine!
	; Input
	;       DS:SI   Null terminated string to print
	;       AH      Row on screen 1-25
	;       AL      Column on screen 1-80
	;       BH      Video Attribute
	; Output
	;       Screen memory (B000:0000 or 8000 CGA/MONO
	; Modifies
	;       SI - Points 1 byte beyond null of str displayed
		push    ax              ; Store the registers
		push    bx
		push    es
		push    di
		call    MoveXY_DI       ; Set DI to row / column offset
		mov     es, [BaseOfScreen]      ; Move screen seg to ES
		cld                     ; Clear the direction flag
		mov     ah, bh          ; Place video attr in AH
	DWS_NoWait:
		lodsb                   ; Get a character from string
		or      al, al          ; Check for NULL
		jz      DWS_Exit        ; If NULL then exit
		stosw                   ; Move word to screen
                jmp     DWS_NoWait      ; Continue
	DWS_Exit:
		pop     di              ; Restore the registers
		pop     es
		pop     bx
		pop     ax
		ret
DWriteStr       ENDP

                PUBLIC StoreToMem
StoreToMem      PROC C
	; This procedure moves an image from the screen to the designated
	; memory area.
        ; *** IMPORTANT -- CALL InitVideo Before using this routine!
	; Input
	;       AH      = Top Row
	;       AL      = Left Column
	;       BH      = Number of rows
	;       BL      = Number of columns
	;       ES:DI   = Memory Destination
	; Modifies
	;       DI
		push    ax
		push    bx
		push    cx
		push    dx
		push    ds              ; Store registers
		push    si
                mov     ds, [BaseOfScreen]      ; Move screen seg to DS
		cld                     ; Clear the direction flag
		mov     ch, 0           ; Clear CH
	STM_NoWait:
		mov     cl, bl          ; Load the number of columns
		call    MoveXY_SI       ; Set SI to screen offset pos
		rep     movsw           ; Thats it!
		inc     ah              ; Next row
		dec     bh              ; Decrement number of rows done
		or      bh, bh          ; Check number of columns
		jnz     STM_NoWait      ; Do next column
		pop     si
		pop     ds
		pop     dx
		pop     cx
		pop     bx
		pop     ax
		ret
StoreToMem      ENDP

                PUBLIC StoreToScr
StoreToScr      PROC C
	; This procedure moves an image from memory to the designated
	; screen location.
        ; *** IMPORTANT -- CALL InitVideo Before using this routine!
	; Input
	;       AH      = Top Row
	;       AL      = Left Column
	;       BH      = Number of rows
	;       BL      = Number of columns
        ;       DX      = Base of segment where image is stored
	;       SI      = Offset from DX to Memory Area of image
	; Modifies
	;       SI
		push    ax                      ; Store Registers
		push    bx
                push    cx
		push    dx
                push    ds
		push    es
		push    di
		mov     es, [BaseOfScreen]      ; Move screen seg to ES
                mov     ds,dx           ; set up new data segment
		cld                     ; Clear the direction flag
		mov     ch, 0           ; Clear CH
	STS_NoWait:
		mov     cl, bl          ; Load the number of columns
		call    MoveXY_DI       ; Set DI to screen offset pos
		rep     movsw           ; Thats it!
		inc     ah              ; Next row
		dec     bh              ; Decrement number of rows done
		or      bh, bh          ; Check number of columns
		jnz     STS_NoWait      ; Do next column
		pop     di              ; Restore registers
		pop     es
                pop     ds
		pop     dx
		pop     cx
		pop     bx
		pop     ax
		ret
StoreToScr      ENDP

                PUBLIC DWriteCHVert
DWriteCHVert    PROC C
        ; Write a character vertically using direct memory access
        ; Input
        ;       AH      Row on screen 1-25
        ;       AL      Column on screen 1-80
        ;       BH      Video Attribute
        ;       BL      Character
        ;       CX      Number of times
                push    ax                ; Store the registers
                push    bx
                push    cx
                push    es
                push    di
                mov     es, [BaseOfScreen]        ; Move screen seg to ES
        DWCV_Loop:
                call    MoveXY_DI       ; Get screen offset in DI
                push    ax              ; Save current row/column
                mov     ax,bx           ; Move char/attr into AX
                stosw                   ; Move char/attr into screen
                pop     ax              ; Restore char/attr into AX
                inc     ah              ; Bump up to next row
                loop    DWCV_Loop       ; Loop back into CX = 0
                pop     di              ; Restore the registers
                pop     es
                pop     cx
                pop     dx
                pop     ax
                ret
DWriteCHVert    ENDP
                END                     ; Of video.asm
