#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys\stat.h>
#include <fcntl.h>
#include <io.h>
#include <mem.h>
#include <dir.h>
#include "video.h"

#define BLOCKSIZE   32000    /* size of buffer block */
#define MAXFIN      500      /* max number of filenames for pop-up */
#define MAXCOL      919      /* max char per line */
#define MAXEDIT     40       /* max width for popup edit */
#define MAXPREV     10       /* max prev file list */
#define POPLEN      19       /* number of filenames per pop-up */
#define TAB         9        /* ASCII tab character */
#define FIRSTFIND   0        /* first find flag */
#define NEXTFIND    1        /* next find flag */
#define NORMFIND    2        /* normal find flag */
#define CASEFIND    3        /* case-sensitive find flag */
#define SUBDIR      16       /* sub-dir attribute flag */
#define FILESA      0        /* file attribute flag */
#define CRKEY       13
#define ESCKEY      27
#define LEFTKEY     331
#define RIGHTKEY    333
#define CTRLLEFT    371
#define UPKEY       328
#define DOWNKEY     336
#define HOMEKEY     327
#define ENDKEY      335
#define PGUPKEY     329
#define PGDNKEY     337
#define F1KEY       315
#define F2KEY       316
#define F5KEY       319
#define F10KEY      324
#define BSKEY       8
#define DELKEY      339
#define CTRLC       3
#define CTRLBREAK   256

#define ANYKEYMSG   "Press ANY key to continue"
#define NOMEMMSG    "Out of memory"
#define NOFILEMSG   "No file(s)"
#define NOOPENMSG   "File error"
#define EMPTYMSG    "** Invalid Filespec **"
#define NOTEXTMSG   "** Text Not Found **"
#define NOLINEMSG   "** Line Not Found **"
#define INLINEMSG   "** Invalid Line Number **"
#define PRINTMSG    "** Printing File **"
#define BOOKSET     "** Bookmark Set **"
#define KEYSMSG     "<ESC=exit>  <F1=help>   Keys: \033\030\031\032 PgUp/PgDn"

extern unsigned BaseOfScreen;
extern unsigned char winlen, winwid;
unsigned char *buffer, *bufmid, *bufend, *endfile, *cptr, *optr, *next,
              *screenhold, wsmode, ypos, showtab, showcr, invattr, normattr,
              framattr, hilattr, nomore, *fn[MAXFIN], *tmpbuf,
              prevfil[MAXPREV][MAXPATH], srchstr[MAXPATH], stdfil[MAXPATH];
long foffset;
int handle, column;
unsigned line;

void initfile(void);
void fileread(unsigned char *buf, unsigned len);
unsigned char nextchar(void);
unsigned char prevchar(void);
void upline(void);
void uppage(void);
void dnline(void);
void dnpage(void);
void disp_line(int y);
void statuswin(void);
void statusmsg(char *s);
void statline(unsigned l);
void exitwin(void);
void display(char *p, char *s);
void exiterror(char *s);
void restoreline(void);
char upfile(unsigned l);
char dnfile(unsigned l);
void getline(void);
char findtext(char fflag, char sflag);
void printfile(void);
void help(void);
void status(char *p, char *s, unsigned b);
char editpopup(char *p, char *s);
char editstring(char *str);
void prevdir(char *p);
int dirlist(char *p);
void sortfn(int num);
void popwin(char plen);
char fileselect(char *p, char *s);
void addprev(char *s);
char selectprev(char *s);
void checkspec(char *p);

int main(int argc, char *argv[])
{
    char p[MAXPATH], s[MAXPATH];

    if(argv[1][1] == '?')
    {
        fprintf(stderr,
            "XBROWSE  Version 2.6  \026  Copyright (c)1994  Hung Vu\n" \
            "Usage: xbrowse [<filename> | <filespec>]\n");
        return 0;
    }
    if(!init_video())
    {
        fprintf(stderr, "Invalid video mode");
        return 0;
    }
    normattr = LIGHTGRAY+BLACK_BG;
    invattr = BLACK+LIGHTGRAY_BG;
    framattr = RED+BLACK_BG;
    hilattr = WHITE+BLACK_BG;
    ypos = wherey();
    tmpbuf = (unsigned char *)malloc(1000);
    if(!tmpbuf) exiterror(NOMEMMSG);
    if(argc < 2)
    {
        if(!strcpy(s, getenv("TEMP"))) strcpy(s, "");
        strcat(s, "\\STDIN.TMP");
        if((handle = creat(s, S_IWRITE)) == -1) exiterror("STDIN error");
        while(gets(tmpbuf))
        {
            strcat(tmpbuf, "\n");
            write(handle, tmpbuf, strlen(tmpbuf));
        }
        close(handle);
        strcpy(stdfil, s);
        strcpy(p, ".");
        checkspec(p);
    }
    else
    {
        strupr(argv[1]);
        strcpy(p, argv[1]);
        checkspec(p);
        strcpy(s, p);
    }
    handle = _open(s, O_RDONLY);
    if(handle == -1)
    {
        if(!fileselect(p, s)) exiterror(NOFILEMSG);
        handle = _open(s, O_RDONLY);
        if(handle == -1) exiterror(NOOPENMSG);
    }
    buffer = (unsigned char *)malloc(BLOCKSIZE);
    if(!buffer) exiterror(NOMEMMSG);
    bufmid = buffer+(BLOCKSIZE/2);
    bufend = bufmid+(BLOCKSIZE/2);
    initfile();
    screenhold = (unsigned char *)malloc((winwid*(winlen+1)*2)+16);
    if(!screenhold) exiterror(NOMEMMSG);
    gettext(1, 1, winwid, (winlen+1), screenhold);
    wsmode = 0xff;
    hidecursor();
    display(p, s);
    close(handle);
    if(stdfil[0]) remove(stdfil);
    exitwin();
    return 0;
}

void initfile(void)
{
    foffset = 0L;
    fileread(buffer, BLOCKSIZE);
    cptr = buffer;
}

void fileread(unsigned char *buf, unsigned len)
{
    int ret;

    endfile = NULL;
    lseek(handle, foffset, SEEK_SET);
    ret = _read(handle, buf, len);
    if(ret == len) return;
    if(ret == -1) ret = 0;
    endfile = buf+ret;
}

unsigned char nextchar(void)
{
    unsigned char ch;

    if(cptr == endfile)
    {
        nomore = 1;
        return *cptr;
    }
    if(cptr == bufend)
    {
        memmove(buffer, bufmid, BLOCKSIZE/2);
        optr = optr-(BLOCKSIZE/2);
        next = next-(BLOCKSIZE/2);
        foffset += (long)BLOCKSIZE;
        fileread(bufmid, BLOCKSIZE/2);
        cptr = bufmid;
        foffset -= (long)(BLOCKSIZE/2);
    }
    nomore = 0;
    ch = *cptr;
    cptr++;
    return ch;
}

unsigned char prevchar(void)
{
    if(cptr == buffer)
    {
        if(foffset == 0L)
        {
            nomore = 1;
            return *cptr;
        }
        memmove(bufmid, buffer, BLOCKSIZE/2);
        optr = optr+(BLOCKSIZE/2);
        next = next+(BLOCKSIZE/2);
        cptr = bufmid;
        foffset -= (long)(BLOCKSIZE/2);
        fileread(buffer, BLOCKSIZE/2);
    }
    nomore = 0;
    cptr--;
    return *cptr;
}

void upline(void)
{
    unsigned char ch;
    
    ch = prevchar();
    if(nomore) return;
    for(;;)
    {
        ch = prevchar();
        if(nomore) break;
        if(ch == '\n')
        {
            nextchar();
            break;
        }
    }
}

void uppage(void)
{
    int i;

    for(i = 0; i < winlen; i++)
    {
        upline();
        if(nomore) break;
        line--;
    }
}

void dnline(void)
{
    unsigned char ch;
    int c = 0;

    for(;;)
    {
        ch = nextchar();
        if((nomore && c == 0) || ch == '\n') break;
        if(nomore && c > 0)
        {
            nomore = 0;
            break;
        }
        c++;
    }
}

void dnpage(void)
{
    int i;

    for(i = 0; i < winlen; i++)
    {
        dnline();
        if(nomore) break;
        line++;
    }
}

void disp_line(int y)
{
    unsigned margin;

    asm push es
    asm mov  di,0
    _CX = y;
    asm mov  ax,160
    asm mul  cx
    asm add  di,ax
    asm mov  ax,BaseOfScreen
    asm mov  es,ax
    margin = winwid+column;
    asm sub  bx,bx
lineloop:
    asm push es
    asm push di
    asm push bx
    _AL = nextchar();
    asm pop  bx
    asm pop  di
    asm pop  es
    asm cmp  nomore,1
    asm je   endline
    asm and  al,wsmode
    asm cmp  al,10
    asm je   endline
    asm cmp  al,13
    asm je   checkcr
    asm cmp  al,9
    asm je   expandtab
showchar:
    asm mov  cx,1
printchar:
    asm cmp  bx,column
    asm jb   noprint
    asm cmp  bx,margin
    asm jae  noprint
    asm mov  ah,normattr
    asm stosw
noprint:
    asm inc  bx
    asm loop printchar
    asm jmp  lineloop
expandtab:
    asm cmp  showtab,1
    asm je   showchar
    asm mov  ax,bx
    asm and  ax,07h
    asm mov  cx,8
    asm sub  cx,ax
    asm mov  al,' '
    asm jmp  printchar
checkcr:
    asm cmp  showcr,1
    asm je   showchar
    asm jmp  lineloop
endline:
    _BX = winwid;
    _CX = y;
    asm mov  ax,160
    asm mul  cx
    asm shl  bx,1
    asm add  bx,ax
    asm mov  cx,bx
    asm sub  cx,di
    asm shr  cx,1
    asm mov  al,' '
    asm mov  ah,normattr
    asm rep  stosw
    asm pop  es
}

void statuswin(void)
{
    window(1, (winlen+1), winwid, (winlen+1));
    textattr(invattr);
}

void statusmsg(char *s)
{
    gotoxy(1, 1);
    cprintf(" XBROWSE V2.6   %-51s", s);
}

void statline(unsigned l)
{
    gotoxy(69, 1);
    cprintf("L%05u:C%03d ", l+1, column+1);
}

void exitwin(void)
{
    if(screenhold) puttext(1, 1, winwid, (winlen+1), screenhold);
    window(1, 1, winwid, (winlen+1));
    textattr(normattr);
    gotoxy(1, ypos);
    showcursor();
}

void display(char *p, char *s)
{
    char str[MAXPATH], udscr, done, fflag, udstw;
    unsigned i, totline, key, bookmark;
    
    line = 0;
    totline = 0;
    column = 0;
    done = 0;
    fflag = NORMFIND;
    bookmark = 0;
    udscr = 1;
    udstw = 0;
    statuswin();
    clrscr();
    statusmsg(KEYSMSG);
    do
    {
        if(udscr)
        {
            optr = cptr;
            for(i = 0; i < winlen; i++) disp_line(i);
            next = cptr;
            cptr = optr;
        }
        statline(line);
        if(!totline && next == endfile)
        {
            optr = cptr;
            i = line;
            dnpage();
            totline = line;
            line = i;
            cptr = optr;
        }
        key = getkey();
        if(udstw) statusmsg(KEYSMSG);
        udstw = 0;
        udscr = 0;
        switch(key)
        {
            case HOMEKEY :
                if(foffset > 0L) initfile();
                if(cptr != buffer || line != 0)
                {
                    cptr = buffer;
                    line = 0;
                    udscr = 1;
                }
                break;
            case ENDKEY :
                if(!totline || line < (totline-1))
                {
                    for(;;)
                    {
                        dnline();
                        if(nomore) break;
                        line++;
                    }
                    upline();
                    line--;
                    udscr = 1;
                }
                break;
            case LEFTKEY :
                if(column > 0)
                {
                    if(column == 4) column -= 4;
                    else column -= 5;
                    if(column < 0) column = 0;
                    udscr = 1;
                }
                break;
            case RIGHTKEY :
                if(column < MAXCOL)
                {
                    if(column == 0) column += 4;
                    else column += 5;
                    if(column > MAXCOL) column = MAXCOL;
                    udscr = 1;
                }
                break;
            case CTRLLEFT :
                if(column > 0)
                {
                    column = 0;
                    udscr = 1;
                }
                break;
            case PGUPKEY :
                if(line > 0)
                {
                    uppage();
                    if(nomore) line = 0;
                    udscr = 1;
                }
                break;
            case PGDNKEY :
                if(!totline || line < (totline-1))
                {
                    cptr = next;
                    line += winlen;
                    if(totline && line > totline) line = totline;
                    if(cptr == endfile)
                    {
                        upline();
                        line--;
                    }
                    udscr = 1;
                }
                break;
            case UPKEY :
                if(line > 0)
                {
                    scroll(DOWN, 0, 0, (winwid-1), (winlen-1), normattr, 1);
                    upline();
                    line--;
                    if(nomore) line = 0;
                    optr = cptr;
                    disp_line(0);
                    if(!totline || (totline-line) > winlen)
                    {
                        cptr = next;
                        upline();
                        next = cptr;
                    }
                    cptr = optr;
                }
                break;
            case DOWNKEY :
                if(!totline || line < (totline-1))
                {
                    scroll(UP, 0, 0, (winwid-1), (winlen-1), normattr, 1);
                    optr = cptr;
                    cptr = next;
                    line++;
                    disp_line((winlen-1));
                    next = cptr;
                    cptr = optr;
                    dnline();
                }
                break;
            case F10KEY :
                strcpy(str, s);
                done = fileselect(p, s);
                if(done && strcmp(str, s))
                {
                    addprev(str);
                    close(handle);
                    handle = _open(s, O_RDONLY);
                    if(handle == -1) exiterror(NOOPENMSG);
                    initfile();
                    line = 0;
                    totline = 0;
                    column = 0;
                    udscr = 1;
                }
                else if(!done)
                {
                    statusmsg(EMPTYMSG);
                    udstw = 1;
                }
                done = 0;
                break;
            case 'g' :
            case 'G' :
                str[0] = 0;
                if(editpopup(" Goto Line ", str))
                {
                    i = atoi(str);
                    if(i > 0 && (!totline || i <= totline))
                    {
                        i--;
                        if(i < line) done = upfile(i);
                        else done = dnfile(i);
                        if(!done)
                        {
                            restoreline();
                            statusmsg(NOLINEMSG);
                        }
                    }
                    else statusmsg(INLINEMSG);
                    udstw = 1;
                    done = 0;
                }
                udscr = 1;
                break;
            case '7' : wsmode = 0x7f; udscr = 1; break;
            case '8' : wsmode = 0xff; udscr = 1; break;
            case 'f' :
            case 'F' :
                fflag = NORMFIND;
                if(!findtext(FIRSTFIND, fflag))
                {
                    statusmsg(NOTEXTMSG);
                    udstw = 1;
                }
                udscr = 1;
                break;
            case 'c' :
            case 'C' :
                fflag = CASEFIND;
                if(!findtext(FIRSTFIND, fflag))
                {
                    statusmsg(NOTEXTMSG);
                    udstw = 1;
                }
                udscr = 1;
                break;
            case 'n' :
            case 'N' :
                if(!findtext(NEXTFIND, fflag))
                {
                    statusmsg(NOTEXTMSG);
                    udstw = 1;
                }
                udscr = 1;
                break;
            case 'p' :
            case 'P' : printfile(); udscr = 1; break;
            case 's' :
            case 'S' :
                editpopup(" Edit Filespec ", p);
                strupr(p);
                checkspec(p);
                udscr = 1;
                break;
            case 't' :
            case 'T' : showtab = !showtab; udscr = 1; break;
            case 'r' :
            case 'R' : showcr = !showcr; udscr = 1; break;
            case F1KEY : help(); udscr = 1; break;
            case F2KEY : status(p, s, bookmark); udscr = 1; break;
            case 'b' :
            case 'B' :
                bookmark = line;
                statusmsg(BOOKSET);
                udstw = 1;
                break;
            case 'j' :
            case 'J' :
                if(bookmark > 0)
                {
                    if(bookmark < line) upfile(bookmark);
                    else dnfile(bookmark);
                    udscr = 1;
                }
                break;
            case F5KEY :
                if(prevfil[0][0])
                {
                    done = selectprev(s);
                    if(done)
                    {
                        close(handle);
                        handle = _open(s, O_RDONLY);
                        if(handle == -1) exiterror(NOOPENMSG);
                        initfile();
                        line = 0;
                        totline = 0;
                        column = 0;
                    }
                    udscr = 1;
                    done = 0;
                }
                break;
            case ESCKEY :
            case CTRLC :
            case CTRLBREAK :
                done = 1;
                break;
        }
    } while(!done);
}

void exiterror(char *s)
{
    close(handle);
    if(stdfil[0]) remove(stdfil);
    exitwin();
    cputs(s);
    exit(1);
}

void restoreline(void)
{
    unsigned i;

    initfile();
    i = line;
    line = 0;
    dnfile(i);
}

char upfile(unsigned l)
{
    unsigned i;

    i = line;
    for(;;)
    {
        if(i == l)
        {
            line = i;
            return 1;
        }
        upline();
        i--;
        if(nomore)
        {
            if(i == l)
            {
                line = i;
                return 1;
            }
            break;
        }
        statline(i);
    }
    return 0;
}

char dnfile(unsigned l)
{
    unsigned i;

    i = line;
    for(;; i++)
    {
        dnline();
        if(nomore) break;
        if(i == l)
        {
            upline();
            line = i;
            return 1;
        }
        statline(i);
    }
    return 0;
}

void getline(void)
{
    asm mov  di,tmpbuf
    asm sub  bx,bx
getloop:
    asm push bx
    asm push di
    _AL = nextchar();
    asm pop  di
    asm pop  bx
    asm cmp  nomore,1
    asm je   getnomore
    asm cmp  al,10
    asm je   getendline
    asm cmp  al,9
    asm je   gettab
getone:
    asm mov  cx,1
getstore:
    asm stosb
    asm inc  bx
    asm cmp  bx,999
    asm ja   getmaxlen
    asm loop getstore
    asm jmp  getloop
getmaxlen:
    asm push bx
    asm push di
    dnline();
    asm pop  di
    asm pop  bx
    asm jmp  getendline
gettab:
    asm cmp  showtab,1
    asm je   getone
    asm mov  ax,bx
    asm and  ax,07h
    asm mov  cx,8
    asm sub  cx,ax
    asm mov  al,' '
    asm jmp  getstore
getnomore:
    asm cmp  bx,0
    asm je   getendline
    nomore = 0;
getendline:
    asm mov  al,0
    asm stosb
}

char findtext(char fflag, char sflag)
{
    char *ret, found;
    unsigned i;

    if(fflag == FIRSTFIND)
    {
        if(sflag == NORMFIND) strcpy(tmpbuf, " Find Text ");
        else strcpy(tmpbuf, " Case-sensitive Find ");
        if(!editpopup(tmpbuf, srchstr)) return 1;
        if(!srchstr[0]) return 1;
        if(sflag == NORMFIND) strupr(srchstr);
        initfile();
        i = 0;
    }
    else
    {
        if(!srchstr[0])
        {
            if(!editpopup(" Find Next ", srchstr)) return 1;
            if(!srchstr[0]) return 1;
            if(sflag == NORMFIND) strupr(srchstr);
        }
        dnline();
        i = line+1;
    }
    found = 0;
    for(;; i++)
    {
        getline();
        if(sflag == NORMFIND) strupr(tmpbuf);
        if(nomore) break;
        statline(i);
        ret = strstr(tmpbuf, srchstr);
        if(ret)
        {
            found = 1;
            column = ret-tmpbuf;
            break;
        }
    }
    if(found)
    {
        line = i;
        upline();
    }
    else restoreline();
    return found;
}

void printfile(void)
{
    unsigned i;
    int ret;

    statusmsg(PRINTMSG);
    initfile();
    i = 0;
    for(;; i++)
    {
        getline();
        statline(i);
        ret = fprintf(stdprn, "%s\r\n", tmpbuf);
        if(ret == -1) break;
        if(nomore)
        {
            fprintf(stdprn, "\f");
            break;
        }
    }
    restoreline();
    statusmsg(KEYSMSG);
}

void help(void)
{
    window(17, 8, 62, 16);
    textattr(invattr);
    framewin(invattr);
    wtitle(" XBROWSE - VERSION 2.6 ");
    clrscr();
    gotoxy(3, 2);
    cputs("G  Goto Line        J  Jump Bookmark");
    gotoxy(3, 3);
    cputs("B  Set Bookmark     C  Case-sensitive Find");
    gotoxy(3, 4);
    cputs("F  Find Text      7/8  Strip/Leave Hi-bit");
    gotoxy(3, 5);
    cputs("N  Next Find      F10  Select New File");
    gotoxy(3, 6);
    cputs("P  Print File       S  Change Filespec");
    gotoxy(3, 7);
    cputs("T  Show/Hide TAB   F2  Current Status");
    gotoxy(3, 8);
    cputs("R  Show/Hide CR    F5  Previous Files");
    statuswin();
    statusmsg(ANYKEYMSG);
    getkey();
    statusmsg(KEYSMSG);
}

void status(char *p, char *s, unsigned b)
{
    window(6, 14, 74, 17);
    textattr(invattr);
    framewin(invattr);
    wtitle(" Status ");
    clrscr();
    gotoxy(3, 1);
    cputs("Current Filespec: ");
    gotoxy(21, 1);
    if(strlen(p) < 48) cprintf("%-48s", p);
    else cprintf("...%-45s", &p[strlen(p)-45]);
    gotoxy(3, 2);
    cputs("Current Filename: ");
    gotoxy(21, 2);
    if(strlen(s) < 48) cprintf("%-48s", s);
    else cprintf("...%-45s", &s[strlen(s)-45]);
    gotoxy(3, 3);
    cprintf("   Searched Text: %-.48s", srchstr);
    gotoxy(3, 4);
    cputs("Current Bookmark: ");
    gotoxy(21, 4);
    if(b > 0) cprintf("%u", b+1);
    statuswin();
    statusmsg(ANYKEYMSG);
    getkey();
    statusmsg(KEYSMSG);
}

char editpopup(char *p, char *s)
{
    char ret;

    window(16, 9, 58, 9);
    framewin(framattr);
    wtitle(p);
    textattr(normattr);
    clrscr();
    cputs("\020 ");
    textattr(hilattr);
    ret = editstring(s);
    statuswin();
    return ret;
}

char editstring(char *str)
{
    char len, pos, xpos, col, ud, done;
    int ch;

    strcpy(tmpbuf, str);
    len = strlen(tmpbuf);
    pos = len;
    col = pos-MAXEDIT;
    if(col < 0) col = 0;
    xpos = pos-col;
    done = 0;
    ud = 1;
    showcursor();
    do
    {
        if(ud)
        {
            hidecursor();
            gotoxy(3, 1);
            cprintf("%-40.40s", &tmpbuf[col]);
            showcursor();
        }
        gotoxy(3+xpos, 1);
        ch = getkey();
        ud = 0;
        switch(ch)
        {
            case HOMEKEY :
                if(pos > 0)
                {
                    pos = 0;
                    xpos = 0;
                    col = 0;
                    ud = 1;
                }
                break;
            case ENDKEY :
                if(pos < len)
                {
                    pos = len;
                    col = pos-MAXEDIT;
                    if(col < 0) col = 0;
                    xpos = pos-col;
                    ud = 1;
                }
                break;
            case LEFTKEY :
                if(pos > 0)
                {
                    pos--;
                    if(xpos > 0) xpos--;
                    else if(col > 0)
                    {
                        col--;
                        ud = 1;
                    }
                }
                break;
            case RIGHTKEY :
                if(pos < len)
                {
                    pos++;
                    if(xpos < MAXEDIT) xpos++;
                    else
                    {
                        col++;
                        ud = 1;
                    }
                }
                break;
            case BSKEY :
                if(pos > 0)
                {
                    movmem(&tmpbuf[pos], &tmpbuf[pos-1], len-pos+1);
                    pos--;
                    len--;
                    if(xpos > 0) xpos--;
                    else
                    {
                        col -= MAXEDIT;
                        if(col < 0) col = 0;
                        xpos = pos-col;
                    }
                    ud = 1;
                }
                break;
            case DELKEY :
                if(pos < len)
                {
                    movmem(&tmpbuf[pos+1], &tmpbuf[pos], len-pos);
                    len--;
                    ud = 1;
                }
                break;
            case CRKEY :
                done = 1;
                break;
            case ESCKEY :
                ch = CRKEY;
                break;
            default :
                if(len < MAXPATH && ch < 255)
                {
                    if(pos < len) memmove(&tmpbuf[pos+1], &tmpbuf[pos], len-pos+1);
                    len++;
                    tmpbuf[pos++] = ch;
                    if(xpos < MAXEDIT) xpos++;
                    else col++;
                    ud = 1;
                }
                break;
        }
        tmpbuf[len] = 0;
    } while(ch != CRKEY);
    hidecursor();
    if(done) strcpy(str, tmpbuf);
    return done;
}

void prevdir(char *p)
{
    char *ret, str[MAXPATH];

    strcpy(str, p);
    if(str[strlen(str)-1] == '\\') str[strlen(str)-1] = 0;
    ret = strrchr(str, '\\');
    if(!ret) return;
    memmove(p, str, (ret-str));
    p[ret-str] = 0;
}

int dirlist(char *p)
{
    struct ffblk fb;
    int done, num;
    char drive[MAXDRIVE], dir[MAXDIR], file[MAXFILE], ext[MAXEXT];
    char sp[MAXPATH];

    fnsplit(p, drive, dir, file, ext);
    strcpy(sp, drive);
    strcat(sp, dir);
    if(sp[strlen(sp)-1] != '\\') strcat(sp, "\\");
    strcat(sp, "*.*");
    num = 0;
    done = findfirst(sp, &fb, SUBDIR);
    if(done) return num;
    if(fb.ff_attrib == SUBDIR && strcmp(fb.ff_name, "."))
    {
        fn[num] = (char *)malloc(strlen(fb.ff_name)+2);
        if(!fn[num]) return num;
        strcpy(fn[num], " ");
        strcat(fn[num], fb.ff_name);
        num++;
    }
    for(;;)
    {
        done = findnext(&fb);
        if(done) break;
        if(fb.ff_attrib != SUBDIR) continue;
        fn[num] = (char *)malloc(strlen(fb.ff_name)+2);
        if(!fn[num]) return num;
        strcpy(fn[num], " ");
        strcat(fn[num], fb.ff_name);
        num++;
        if(num >= MAXFIN) break;
    }
    if(num < MAXFIN)
    {
        done = findfirst(p, &fb, FILESA);
        if(!done)
        {
            fn[num] = (char *)malloc(strlen(fb.ff_name)+1);
            if(!fn[num]) return num;
            strcpy(fn[num], fb.ff_name);
            num++;
            if(num < MAXFIN)
            {
                for(;;)
                {
                    done = findnext(&fb);
                    if(done) break;
                    fn[num] = (char *)malloc(strlen(fb.ff_name)+1);
                    if(!fn[num]) return num;
                    strcpy(fn[num], fb.ff_name);
                    num++;
                    if(num >= MAXFIN) break;
                }
            }
        }
    }
    return num;
}

void sortfn(int num)
{
    char *tmp, done;
    int i;

    do
    {
        done = 0;
        for(i = 1; i < num; i++)
        {
            if(strcmp(fn[i-1], fn[i]) > 0)
            {
                tmp = fn[i-1];
                fn[i-1] = fn[i];
                fn[i] = tmp;
                done = 1;
            }
        }
    } while(done);
    for(i = 0;; i++)
    {
        if(fn[i][0] != ' ') break;
        fn[i][0] = '\\';
    }
}

void popwin(char plen)
{
    gettext(31, 3, 48, plen+4, tmpbuf);
    window(32, 4, 47, plen+3);
    framewin(framattr);
    textattr(normattr);
    clrscr();
}

char fileselect(char *p, char *s)
{
    char str[18], file[MAXFILE], ext[MAXEXT];
    char drive[MAXDRIVE], dir[MAXDIR], oldp[MAXPATH];
    int key, top, i, num, y, plen;
    char udscr, udlin, done;

    num = dirlist(p);
    if(!num) return 0;
    sortfn(num);
    strcpy(oldp, p);
    if(num < POPLEN) plen = num;
    else plen = POPLEN;
    top = 0;
    y = 1;
    fnsplit(s, drive, dir, file, ext);
    strcpy(str, file);
    strcat(str, ext);
    for(i = 0; i < num; i++)
    {
        if(strcmp(str, fn[i]) == 0)
        {
            if(i < (top+plen)) y = i-top+1;
            else
            {
                top = i-plen+1;
                y = plen;
            }
            break;
        }
    }
    done = 0;
    udscr = 1;
    udlin = 1;
    popwin(plen);
    hidecursor();
    do
    {
        if(udscr)
        {
            for(i = 0; i < plen; i++)
            {
                gotoxy(1, i+1);
                if(fn[top+i][0] == '\\')
                {
                    textattr(hilattr);
                    cprintf("  %-14s", fn[top+i]);
                }
                else
                {
                    fnsplit(fn[top+i], drive, dir, file, ext);
                    textattr(normattr);
                    cprintf("  %-8s%-6s", file, ext);
                }
            }
        }
        if(udlin)
        {
            textattr(invattr);
            if(fn[top+y-1][0] == '\\') sprintf(str, "  %-14s", fn[top+y-1]);
            else
            {
                fnsplit(fn[top+y-1], drive, dir, file, ext);
                sprintf(str, "  %-8s%-6s", file, ext);
            }
            gotoxy(1, y);
            cputs(str);
            if(fn[top+y-1][0] == '\\') textattr(hilattr);
            else textattr(normattr);
        }
        key = getkey();
        udscr = 0;
        udlin = 0;
        switch(key)
        {
            case ESCKEY :
                strcpy(p, oldp);
                done = 1;
                break;
            case HOMEKEY :
                if((top+y-1) > 0)
                {
                    top = 0;
                    y = 1;
                    udscr = 1;
                    udlin = 1;
                }
                break;
            case ENDKEY :
                if((top+y) < num)
                {
                    top = num-plen;
                    y = plen;
                    udscr = 1;
                    udlin = 1;
                }
                break;
            case PGUPKEY :
                if((top+y-1) > 0)
                {
                    top = top-(plen-1);
                    if(top < 0) top = 0;
                    y = 1;
                    udscr = 1;
                    udlin = 1;
                }
                break;
            case PGDNKEY :
                if(top < (num-plen))
                {
                    top = top+(plen-1);
                    if(top > num-plen) top = num-plen;
                    y = 1;
                    udscr = 1;
                    udlin = 1;
                }
                break;
            case UPKEY :
                if((top+y-1) > 0)
                {
                    gotoxy(1, y);
                    cputs(str);
                    if(y == 1)
                    {
                        scroll(DOWN, 33, 3, 46, plen+2, normattr, 1);
                        top--;
                    }
                    else y--;
                    udlin = 1;
                }
                break;
            case DOWNKEY :
                if((top+y) < num)
                {
                    gotoxy(1, y);
                    cputs(str);
                    if(y == plen)
                    {
                        scroll(UP, 33, 3, 46, plen+2, normattr, 1);
                        top++;
                    }
                    else y++;
                    udlin = 1;
                }
                break;
            case CRKEY :
                if(fn[top+y-1][0] != '\\')
                {
                    fnsplit(p, drive, dir, file, ext);
                    strcpy(s, drive);
                    strcat(s, dir);
                    strcat(s, fn[top+y-1]);
                    strupr(s);
                    done = 1;
                }
                else
                {
                    fnsplit(p, drive, dir, file, ext);
                    if(strcmp(fn[top+y-1], "\\..") == 0) prevdir(dir);
                    else strcat(dir, &fn[top+y-1][1]);
                    strcpy(p, drive);
                    strcat(p, dir);
                    strcat(p, "\\");
                    strcat(p, file);
                    strcat(p, ext);
                    for(i = 0; i < num; i++) if(fn[i]) free(fn[i]);
                    num = dirlist(p);
                    if(!num) done = 1;
                    else
                    {
                        sortfn(num);
                        puttext(31, 3, 48, plen+4, tmpbuf);
                        if(num < POPLEN) plen = num;
                        else plen = POPLEN;
                        top = 0;
                        done = 0;
                        y = 1;
                        udscr = 1;
                        udlin = 1;
                        popwin(plen);
                    }
                }
                break;
            default :
                if(islower(key)) key = _toupper(key);
                for(i = (top+y); i < num; i++)
                {
                    if(fn[i][0] == key)
                    {
                        gotoxy(1, y);
                        cputs(str);
                        if(i < (top+plen))
                        {
                            y = i-top+1;
                            udlin = 1;
                        }
                        else
                        {
                            top = i-plen+1;
                            y = plen;
                            udscr = 1;
                            udlin = 1;
                        }
                        break;
                    }
                }
                break;
        }
    } while(!done);
    puttext(31, 3, 48, plen+4, tmpbuf);
    for(i = 0; i < num; i++) if(fn[i]) free(fn[i]);
    statuswin();
    return done;
}

void addprev(char *s)
{
    char i;

    for(i = 0; i < MAXPREV; i++) if(!prevfil[i][0]) break;
    if(i >= MAXPREV)
    {
        for(i = 0; i < (MAXPREV-1); i++) strcpy(prevfil[i], prevfil[i+1]);
    }
    strcpy(prevfil[i], s);
}

char selectprev(char *s)
{
    char i, y, done, maxp, ud;
    int key;

    for(i = 0; i < MAXPREV; i++) if(!prevfil[i][0]) break;
    maxp = i;
    window(20, 6, 60, maxp+5);
    framewin(framattr);
    textattr(normattr);
    y = 1;
    for(i = 0; i < maxp; i++)
    {
        gotoxy(1, y+i);
        if(strlen(prevfil[i]) < 39) cprintf(" %-40s", prevfil[i]);
        else cprintf(" ...%-37s", &prevfil[i][strlen(prevfil[i])-36]);
    }
    done = 0;
    ud = 1;
    do
    {
        if(ud)
        {
            textattr(invattr);
            if(strlen(prevfil[y-1]) < 39) sprintf(tmpbuf, " %-40s", prevfil[y-1]);
            else sprintf(tmpbuf, " ...%-37s", &prevfil[y-1][strlen(prevfil[y-1])-36]);
            gotoxy(1, y);
            cputs(tmpbuf);
            textattr(normattr);
        }
        key = getkey();
        ud = 0;
        switch(key)
        {
            case UPKEY :
                if(y > 1)
                {
                    gotoxy(1, y);
                    cputs(tmpbuf);
                    y--;
                    ud = 1;
                }
                break;
            case DOWNKEY :
                if(y < maxp)
                {
                    gotoxy(1, y);
                    cputs(tmpbuf);
                    y++;
                    ud = 1;
                }
                break;
            case ESCKEY :
                key = CRKEY;
                break;
            case CRKEY :
                strcpy(tmpbuf, s);
                strcpy(s, prevfil[y-1]);
                addprev(tmpbuf);
                done = 1;
                break;
        }
    } while(key != CRKEY);
    statuswin();
    return done;
}

void checkspec(char *p)
{
    struct ffblk fb;
    char done;

    strcpy(tmpbuf, p);
    if(p[0] == '.')
    {
        getcwd(tmpbuf, MAXPATH);
        if(p[1] == '.') prevdir(tmpbuf);
        if(tmpbuf[strlen(tmpbuf)-1] != '\\') strcat(tmpbuf, "\\");
        if(strlen(p) > 2)
        {
            if(p[1] == '\\') strcat(tmpbuf, &p[2]);
            else if(p[2] == '\\') strcat(tmpbuf, &p[3]);
        }
    }
    else if(p[0] != '\\' && p[1] != ':')
    {
        getcwd(tmpbuf, MAXPATH);
        if(tmpbuf[strlen(tmpbuf)-1] != '\\') strcat(tmpbuf, "\\");
        strcat(tmpbuf, p);
    }
    if(tmpbuf[strlen(tmpbuf)-1] == '\\') strcat(tmpbuf, "*.*");
    done = findfirst(tmpbuf, &fb, FILESA);
    if(done)
    {
        done = findfirst(tmpbuf, &fb, SUBDIR);
        if(!done && fb.ff_name[0] != '.' && !strrchr(tmpbuf, '*')) strcat(tmpbuf, "\\*.*");
    }
    strcpy(p, tmpbuf);
}
